# Copyright 2015, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities related to the PowerVM management partition.

The management partition is a special LPAR that runs the PowerVM REST API
service.  It itself appears through the REST API as a LogicalPartition of type
aixlinux, but with the is_mgmt_partition property set to True.

The PowerVM Nova Compute service runs on the management partition.
"""
import glob
from nova import exception
from nova import utils
import os
from os import path
from oslo_concurrency import lockutils
from oslo_log import log as logging
from pypowervm.tasks import partition as pvm_par
import retrying

from nova_powervm.virt.powervm import exception as npvmex


LOG = logging.getLogger(__name__)

_MP_UUID = None


@lockutils.synchronized("mgmt_lpar_uuid")
def mgmt_uuid(adapter):
    """Returns the management partitions UUID."""
    global _MP_UUID
    if not _MP_UUID:
        _MP_UUID = pvm_par.get_this_partition(adapter).uuid
    return _MP_UUID


def _tee_as_root(fpath, payload):
    """Executes 'echo $payload | sudo tee -a $fpath'.

    :param fpath: The file system path to which to tee.
    :param payload: The string to pipe to the file.
    """
    utils.execute('tee', '-a', fpath, process_input=payload, run_as_root=True)


def discover_vscsi_disk(mapping, scan_timeout=300):
    """Bring a mapped device into the management partition and find its name.

    Based on a VSCSIMapping, scan the appropriate virtual SCSI host bus,
    causing the operating system to discover the mapped device.  Find and
    return the path of the newly-discovered device based on its UDID in the
    mapping.

    Note: scanning the bus will cause the operating system to discover *all*
    devices on that bus.  However, this method will only return the path for
    the specific device from the input mapping, based on its UDID.

    :param mapping: The pypowervm.wrappers.virtual_io_server.VSCSIMapping
                    representing the mapping of the desired disk to the
                    management partition.
    :param scan_timeout: The maximum number of seconds after scanning to wait
                         for the specified device to appear.
    :return: The udev-generated ("/dev/sdX") name of the discovered disk.
    :raise NoDiskDiscoveryException: If the disk did not appear after the
                                     specified timeout.
    :raise UniqueDiskDiscoveryException: If more than one disk appears with the
                                         expected UDID.
    """
    # TODO(IBM): Support for other host platforms.

    # Calculate the Linux slot number from the client adapter slot number.
    lslot = 0x30000000 | mapping.client_adapter.lpar_slot_num
    # We'll match the device ID based on the UDID, which is actually the last
    # 32 chars of the field we get from PowerVM.
    udid = mapping.backing_storage.udid[-32:]

    LOG.debug("Trying to discover VSCSI disk with UDID %(udid)s on slot "
              "%(slot)x.", {'udid': udid, 'slot': lslot})

    # Find the special file to scan the bus, and scan it.
    # This glob should yield exactly one result, but use the loop just in case.
    for scanpath in glob.glob(
            '/sys/bus/vio/devices/%x/host*/scsi_host/host*/scan' % lslot):
        # echo '- - -' | sudo tee -a /path/to/scan
        _tee_as_root(scanpath, '- - -')

    # Now see if our device showed up.  If so, we can reliably match it based
    # on its Linux ID, which ends with the disk's UDID.
    dpathpat = '/dev/disk/by-id/*%s' % udid

    # The bus scan is asynchronous.  Need to poll, waiting for the device to
    # spring into existence.  Stop when glob finds at least one device, or
    # after the specified timeout.  Sleep 1/4 second between polls.
    @retrying.retry(retry_on_result=lambda result: not result, wait_fixed=250,
                    stop_max_delay=scan_timeout * 1000)
    def _poll_for_dev(globpat):
        return glob.glob(globpat)
    try:
        disks = _poll_for_dev(dpathpat)
    except retrying.RetryError as re:
        raise npvmex.NoDiskDiscoveryException(
            bus=lslot, udid=udid, polls=re.last_attempt.attempt_number,
            timeout=scan_timeout)
    # If we get here, _poll_for_dev returned a nonempty list.  If not exactly
    # one entry, this is an error.
    if len(disks) != 1:
        raise npvmex.UniqueDiskDiscoveryException(path_pattern=dpathpat,
                                                  count=len(disks))

    # The by-id path is a symlink.  Resolve to the /dev/sdX path
    dpath = path.realpath(disks[0])
    LOG.debug("Discovered VSCSI disk with UDID %(udid)s on slot %(slot)x at "
              "path %(devname)s.",
              {'udid': udid, 'slot': lslot, 'devname': dpath})
    return dpath


def remove_block_dev(devpath, scan_timeout=10):
    """Remove a block device from the management partition.

    This method causes the operating system of the management partition to
    delete the device special files associated with the specified block device.

    :param devpath: Any path to the block special file associated with the
                    device to be removed.
    :param scan_timeout: The maximum number of seconds after scanning to wait
                         for the specified device to disappear.
    :raise InvalidDevicePath: If the specified device or its 'delete' special
                              file cannot be found.
    :raise DeviceDeletionException: If the deletion was attempted, but the
                                    device special file is still present
                                    afterward.
    """
    # Resolve symlinks, if any, to get to the /dev/sdX path
    devpath = path.realpath(devpath)
    try:
        os.stat(devpath)
    except OSError:
        raise exception.InvalidDevicePath(path=devpath)
    devname = devpath.rsplit('/', 1)[-1]
    delpath = '/sys/block/%s/device/delete' % devname
    try:
        os.stat(delpath)
    except OSError:
        raise exception.InvalidDevicePath(path=delpath)
    LOG.debug("Deleting block device %(devpath)s from the management "
              "partition via special file %(delpath)s.",
              {'devpath': devpath, 'delpath': delpath})
    _tee_as_root(delpath, '1')

    # The bus scan is asynchronous.  Need to poll, waiting for the device to
    # disappear.  Stop when stat raises OSError (dev file not found) - which is
    # success - or after the specified timeout (which is failure).  Sleep 1/4
    # second between polls.
    @retrying.retry(retry_on_result=lambda result: result, wait_fixed=250,
                    stop_max_delay=scan_timeout * 1000)
    def _poll_for_del(statpath):
        try:
            os.stat(statpath)
            return True
        except OSError:
            # Device special file is absent, as expected
            return False
    try:
        _poll_for_del(devpath)
    except retrying.RetryError as re:
        # stat just kept returning (dev file continued to exist).
        raise npvmex.DeviceDeletionException(
            devpath=devpath, polls=re.last_attempt.attempt_number,
            timeout=scan_timeout)
    # Else stat raised - the device disappeared - all done.



# Copyright 2014, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import mock
from nova.compute import power_state
from nova import exception
from nova import test
from pypowervm.wrappers import event as pvm_evt

from nova_powervm.virt.powervm import event


class TestGetInstance(test.TestCase):
    @mock.patch('nova.context.get_admin_context')
    @mock.patch('nova_powervm.virt.powervm.vm.get_instance')
    def test_get_instance(self, mock_get_inst, mock_get_context):
        # If instance provided, vm.get_instance not called
        self.assertEqual('inst', event._get_instance('inst', 'uuid'))
        mock_get_inst.assert_not_called()
        # Note that we can only guarantee get_admin_context wasn't called
        # because _get_instance is mocked everywhere else in this suite.
        # Otherwise it could run from another test case executing in parallel.
        mock_get_context.assert_not_called()

        # If instance not provided, vm.get_instance is called
        mock_get_inst.return_value = 'inst2'
        for _ in range(2):
            # Doing it the second time doesn't call get_admin_context() again.
            self.assertEqual('inst2', event._get_instance(None, 'uuid'))
            mock_get_context.assert_called_once_with()
            mock_get_inst.assert_called_once_with(
                mock_get_context.return_value, 'uuid')
            mock_get_inst.reset_mock()
            # Don't reset mock_get_context


class TestPowerVMNovaEventHandler(test.TestCase):
    def setUp(self):
        super(TestPowerVMNovaEventHandler, self).setUp()
        lceh_process_p = mock.patch(
            'nova_powervm.virt.powervm.event.PowerVMLifecycleEventHandler.'
            'process')
        self.addCleanup(lceh_process_p.stop)
        self.mock_lceh_process = lceh_process_p.start()
        self.mock_driver = mock.Mock()
        self.handler = event.PowerVMNovaEventHandler(self.mock_driver)

    @mock.patch('nova_powervm.virt.powervm.event._get_instance')
    def test_handle_inst_event(self, mock_get_instance):
        # If no event we care about, or NVRAM but no nvram_mgr, nothing happens
        self.mock_driver.nvram_mgr = None
        for dets in ([], ['foo', 'bar', 'baz'], ['NVRAM']):
            self.assertEqual('inst', self.handler._handle_inst_event(
                'inst', 'uuid', dets))
        mock_get_instance.assert_not_called()
        self.mock_lceh_process.assert_not_called()

        self.mock_driver.nvram_mgr = mock.Mock()

        # PartitionState only: no NVRAM handling, and inst is passed through.
        self.assertEqual('inst', self.handler._handle_inst_event(
            'inst', 'uuid', ['foo', 'PartitionState', 'bar']))
        mock_get_instance.assert_not_called()
        self.mock_driver.nvram_mgr.store.assert_not_called()
        self.mock_lceh_process.assert_called_once_with('inst', 'uuid')

        self.mock_lceh_process.reset_mock()

        # No instance; nothing happens (we skip PartitionState handling too)
        mock_get_instance.return_value = None
        self.assertIsNone(self.handler._handle_inst_event(
            'inst', 'uuid', ['NVRAM', 'PartitionState']))
        mock_get_instance.assert_called_once_with('inst', 'uuid')
        self.mock_driver.nvram_mgr.store.assert_not_called()
        self.mock_lceh_process.assert_not_called()

        mock_get_instance.reset_mock()
        mock_get_instance.return_value = 'inst'

        # NVRAM only - no PartitionState handling, instance is returned
        self.assertEqual('inst', self.handler._handle_inst_event(
            None, 'uuid', ['NVRAM', 'baz']))
        mock_get_instance.assert_called_once_with(None, 'uuid')
        self.mock_driver.nvram_mgr.store.assert_called_once_with('inst')
        self.mock_lceh_process.assert_not_called()

        mock_get_instance.reset_mock()
        self.mock_driver.nvram_mgr.store.reset_mock()

        # Both event types
        self.assertEqual('inst', self.handler._handle_inst_event(
            None, 'uuid', ['PartitionState', 'NVRAM']))
        mock_get_instance.assert_called_once_with(None, 'uuid')
        self.mock_driver.nvram_mgr.store.assert_called_once_with('inst')
        self.mock_lceh_process.assert_called_once_with('inst', 'uuid')

    @mock.patch('nova_powervm.virt.powervm.event.PowerVMNovaEventHandler.'
                '_handle_inst_event')
    @mock.patch('pypowervm.util.get_req_path_uuid', autospec=True)
    def test_process(self, mock_get_rpu, mock_handle):
        # NEW_CLIENT/CACHE_CLEARED events are ignored
        events = [mock.Mock(etype=pvm_evt.EventType.NEW_CLIENT),
                  mock.Mock(etype=pvm_evt.EventType.CACHE_CLEARED)]
        self.handler.process(events)
        self.assertEqual(0, mock_get_rpu.call_count)
        mock_handle.assert_not_called()

        moduri = pvm_evt.EventType.MODIFY_URI
        # If get_req_path_uuid doesn't find a UUID, or not a LogicalPartition
        # URI, or details is empty, or has no actions we care about, no action
        # is taken.
        mock_get_rpu.side_effect = [None, 'uuid1', 'uuid2', 'uuid3']
        events = [
            mock.Mock(etype=moduri, data='foo/LogicalPartition/None',
                      details='NVRAM,PartitionState'),
            mock.Mock(etype=moduri, data='bar/VirtualIOServer/uuid1',
                      details='NVRAM,PartitionState'),
            mock.Mock(etype=moduri, data='baz/LogicalPartition/uuid2',
                      detail=''),
            mock.Mock(etype=moduri, data='blah/LogicalPartition/uuid3',
                      detail='do,not,care')]
        self.handler.process(events)
        mock_get_rpu.assert_has_calls(
            [mock.call(uri, preserve_case=True)
             for uri in ('bar/VirtualIOServer/uuid1',
                         'baz/LogicalPartition/uuid2',
                         'blah/LogicalPartition/uuid3')])
        mock_handle.assert_not_called()

        mock_get_rpu.reset_mock()

        # The stars align, and we handle some events.
        uuid_det = (('uuid1', 'NVRAM'),
                    ('uuid2', 'this,one,ignored'),
                    ('uuid3', 'PartitionState,baz,NVRAM'),
                    # Repeat uuid1 to test the cache
                    ('uuid1', 'blah,PartitionState'),
                    ('uuid5', 'also,ignored'))
        mock_get_rpu.side_effect = [ud[0] for ud in uuid_det]
        events = [
            mock.Mock(etype=moduri, data='LogicalPartition/' + uuid,
                      detail=detail) for uuid, detail in uuid_det]
        # Set up _handle_inst_event to test the cache and the exception path
        mock_handle.side_effect = ['inst1', None, ValueError]
        # Run it!
        self.handler.process(events)
        mock_get_rpu.assert_has_calls(
            [mock.call(uri, preserve_case=True) for uri in
             ('LogicalPartition/' + ud[0] for ud in uuid_det)])
        mock_handle.assert_has_calls(
            [mock.call(None, 'uuid1', ['NVRAM']),
             mock.call(None, 'uuid3', ['PartitionState', 'baz', 'NVRAM']),
             # inst1 pulled from the cache based on uuid1
             mock.call('inst1', 'uuid1', ['blah', 'PartitionState'])])


class TestPowerVMLifecycleEventHandler(test.TestCase):
    def setUp(self):
        super(TestPowerVMLifecycleEventHandler, self).setUp()
        self.mock_driver = mock.MagicMock()
        self.handler = event.PowerVMLifecycleEventHandler(self.mock_driver)

    @mock.patch('nova_powervm.virt.powervm.vm.get_vm_qp')
    @mock.patch('nova_powervm.virt.powervm.event._get_instance')
    @mock.patch('nova_powervm.virt.powervm.vm.translate_event')
    @mock.patch('nova.virt.event.LifecycleEvent')
    def test_emit_event(self, mock_lce, mock_tx_evt, mock_get_inst, mock_qp):
        def assert_qp():
            mock_qp.assert_called_once_with(
                self.mock_driver.adapter, 'uuid', 'PartitionState')
            mock_qp.reset_mock()

        def assert_get_inst():
            mock_get_inst.assert_called_once_with('inst', 'uuid')
            mock_get_inst.reset_mock()

        # Ignore if LPAR is gone
        mock_qp.side_effect = exception.InstanceNotFound(instance_id='uuid')
        self.handler._emit_event('uuid', None)
        assert_qp()
        mock_get_inst.assert_not_called()
        mock_tx_evt.assert_not_called()
        mock_lce.assert_not_called()
        self.mock_driver.emit_event.assert_not_called()

        # Let get_vm_qp return its usual mock from now on
        mock_qp.side_effect = None

        # Ignore if instance is gone
        mock_get_inst.return_value = None
        self.handler._emit_event('uuid', 'inst')
        assert_qp()
        assert_get_inst()
        mock_tx_evt.assert_not_called()
        mock_lce.assert_not_called()
        self.mock_driver.emit_event.assert_not_called()

        # Ignore if task_state isn't one we care about
        for task_state in event._NO_EVENT_TASK_STATES:
            mock_get_inst.return_value = mock.Mock(task_state=task_state)
            self.handler._emit_event('uuid', 'inst')
            assert_qp()
            assert_get_inst()
            mock_tx_evt.assert_not_called()
            mock_lce.assert_not_called()
            self.mock_driver.emit_event.assert_not_called()

        # Task state we care about from now on
        inst = mock.Mock(task_state='scheduling',
                         power_state=power_state.RUNNING)
        mock_get_inst.return_value = inst

        # Ignore if not a transition we care about
        mock_tx_evt.return_value = None
        self.handler._emit_event('uuid', 'inst')
        assert_qp()
        assert_get_inst()
        mock_tx_evt.assert_called_once_with(
            mock_qp.return_value, power_state.RUNNING)
        mock_lce.assert_not_called()
        self.mock_driver.emit_event.assert_not_called()

        mock_tx_evt.reset_mock()

        # Good path
        mock_tx_evt.return_value = 'transition'
        self.handler._delayed_event_threads = {'uuid': 'thread1',
                                               'uuid2': 'thread2'}
        self.handler._emit_event('uuid', 'inst')
        assert_qp()
        assert_get_inst()
        mock_tx_evt.assert_called_once_with(
            mock_qp.return_value, power_state.RUNNING)
        mock_lce.assert_called_once_with(inst.uuid, 'transition')
        self.mock_driver.emit_event.assert_called_once_with(
            mock_lce.return_value)
        # The thread was removed
        self.assertEqual({'uuid2': 'thread2'},
                         self.handler._delayed_event_threads)

    @mock.patch('eventlet.greenthread.spawn_after')
    def test_process(self, mock_spawn):
        thread1 = mock.Mock()
        thread2 = mock.Mock()
        mock_spawn.side_effect = [thread1, thread2]
        # First call populates the delay queue
        self.assertEqual({}, self.handler._delayed_event_threads)
        self.handler.process(None, 'uuid')
        mock_spawn.assert_called_once_with(15, self.handler._emit_event,
                                           'uuid', None)
        self.assertEqual({'uuid': thread1},
                         self.handler._delayed_event_threads)
        thread1.cancel.assert_not_called()
        thread2.cancel.assert_not_called()

        mock_spawn.reset_mock()

        # Second call cancels the first thread and replaces it in delay queue
        self.handler.process('inst', 'uuid')
        mock_spawn.assert_called_once_with(15, self.handler._emit_event,
                                           'uuid', 'inst')
        self.assertEqual({'uuid': thread2},
                         self.handler._delayed_event_threads)
        thread1.cancel.assert_called_once_with()
        thread2.cancel.assert_not_called()



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.virt import imagecache

from nova_powervm.virt.powervm.disk import driver

from oslo_log import log as logging
from pypowervm.tasks import storage as tsk_stg
from pypowervm.wrappers import storage as pvm_stg
from pypowervm.wrappers import virtual_io_server as pvm_vios


LOG = logging.getLogger(__name__)


class ImageManager(imagecache.ImageCacheManager):

    def __init__(self, vios_uuid, vg_uuid, adapter):
        super(ImageManager, self).__init__()
        self.vios_uuid = vios_uuid
        self.vg_uuid = vg_uuid
        self.adapter = adapter

    def _get_base(self):
        """Returns the base directory of the cached images.

        :return: Volume Group containing all images/instances
        """
        # Return VG for instances
        return pvm_stg.VG.get(
            self.adapter, uuid=self.vg_uuid,
            parent_type=pvm_vios.VIOS.schema_type, parent_uuid=self.vios_uuid)

    def _scan_base_image(self, base_dir):
        """Scan base images present in base_dir.

        :param base_dir: Volume group containing all images/instances
        :return: List of all virtual disks containing a bootable image that
                 were created for caching purposes.
        """
        # Find LVs in the _get_base VG with i_<partial id>
        prefix = '%s_' % driver.DiskType.IMAGE[0]
        return [image for image in base_dir.virtual_disks
                if image.name.startswith(prefix)]

    def _age_and_verify_cached_images(self, context, all_instances, base_dir):
        """Finds and removes unused images from the cache.

        :param context: nova context
        :param all_instances: List of all instances on the node
        :param base_dir: Volume group of cached images
        """
        # Use the 'used_images' key from nova imagecache to get a dict that
        # uses image_ids as keys.
        cache = self._scan_base_image(base_dir)
        running_inst = self._list_running_instances(context, all_instances)
        adjusted_ids = []
        for img_id in running_inst.get('used_images'):
            if img_id:
                adjusted_ids.append(
                    driver.DiskAdapter.get_name_by_uuid(driver.DiskType.IMAGE,
                                                        img_id, short=True))
        # Compare base images with running instances remove unused
        unused = [image for image in cache if image.name not in adjusted_ids]
        # Remove unused
        if unused:
            for image in unused:
                LOG.info("Removing unused cache image: '%s'", image.name)
            tsk_stg.rm_vg_storage(base_dir, vdisks=unused)

    def update(self, context, all_instances):
        """Remove cached images not being used by any instance.

        :param context: nova context
        :param all_instances: List of all instances on the node
        """

        base_dir = self._get_base()
        self._age_and_verify_cached_images(context, all_instances, base_dir)



# Copyright 2015, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from nova_powervm import conf as cfg
from nova_powervm.tests.virt.powervm.volume import test_driver as test_vol
from nova_powervm.virt.powervm.volume import iscsi

from pypowervm import const as pvm_const
from pypowervm.tasks import hdisk
from pypowervm.tests.tasks.util import load_file
from pypowervm.tests import test_fixtures as pvm_fx
from pypowervm.wrappers import storage as pvm_stor
from pypowervm.wrappers import virtual_io_server as pvm_vios

CONF = cfg.CONF

VIOS_FEED = 'fake_vios_feed2.txt'


class TestISCSIAdapter(test_vol.TestVolumeAdapter):
    """Tests the vSCSI Volume Connector Adapter.  Single VIOS tests"""

    def setUp(self):
        super(TestISCSIAdapter, self).setUp()
        self.adpt = self.useFixture(pvm_fx.AdapterFx()).adpt

        self.vios_feed_resp = load_file(VIOS_FEED)

        self.feed = pvm_vios.VIOS.wrap(self.vios_feed_resp)
        self.ft_fx = pvm_fx.FeedTaskFx(self.feed)
        self.useFixture(self.ft_fx)

        self.adpt.read.return_value = self.vios_feed_resp

        @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS', autospec=True)
        @mock.patch('nova_powervm.virt.powervm.vm.get_pvm_uuid')
        @mock.patch('pypowervm.tasks.partition.get_mgmt_partition',
                    autospec=True)
        @mock.patch('pypowervm.tasks.hdisk.discover_iscsi_initiator',
                    autospec=True)
        def init_vol_adpt(mock_initiator, mock_mgmt_part, mock_pvm_uuid,
                          mock_vios):
            self.trans_type = 'iscsi'
            self.iqn = 'iqn.2016-08.bar.foo:target'
            self.lun = '1'
            self.host_ip = '10.0.0.1'
            self.user = 'user'
            self.password = 'password'
            self.serial = 'f042c68a-c5a5-476a-ba34-2f6d43f4226c'
            con_info = {
                'serial': self.serial,
                'driver_volume_type': self.trans_type,
                'data': {
                    'target_iqn': self.iqn,
                    'target_lun': self.lun,
                    'target_portal': self.host_ip,
                    'auth_username': self.user,
                    'auth_password': self.password
                },
            }
            mock_inst = mock.MagicMock()
            mock_pvm_uuid.return_value = '1234'
            mock_initiator.return_value = 'initiatior iqn'
            # The getter can just return the VIOS values (to remove a read
            # that would otherwise need to be mocked).
            mock_vios.getter.return_value = self.feed
            return iscsi.IscsiVolumeAdapter(self.adpt, 'host_uuid', mock_inst,
                                            con_info)
        self.vol_drv = init_vol_adpt()

        # setup system_metadata tests
        self.devname = "/dev/fake"
        self.slot_mgr = mock.Mock()
        self.slot_mgr.build_map.get_vscsi_slot.return_value = 62, 'the_lua'

    @mock.patch('pypowervm.tasks.hdisk.discover_iscsi', autospec=True)
    @mock.patch('pypowervm.tasks.scsi_mapper.add_map', autospec=True)
    @mock.patch('pypowervm.tasks.scsi_mapper.build_vscsi_mapping',
                autospec=True)
    @mock.patch('pypowervm.tasks.hdisk.lua_recovery', autospec=True)
    @mock.patch('nova_powervm.virt.powervm.vm.get_vm_id')
    def test_connect_volume(self, mock_get_vm_id, mock_lua_recovery,
                            mock_build_map, mock_add_map, mock_discover):
        # The mock return values
        mock_lua_recovery.return_value = (
            hdisk.LUAStatus.DEVICE_AVAILABLE, 'devname', 'udid')
        mock_get_vm_id.return_value = '2'
        mock_discover.return_value = '/dev/fake', 'fake_udid'

        def build_map_func(host_uuid, vios_w, lpar_uuid, pv,
                           lpar_slot_num=None, lua=None, target_name=None):
            self.assertEqual('host_uuid', host_uuid)
            self.assertIsInstance(vios_w, pvm_vios.VIOS)
            self.assertEqual('1234', lpar_uuid)
            self.assertIsInstance(pv, pvm_stor.PV)
            self.assertEqual(62, lpar_slot_num)
            self.assertEqual('the_lua', lua)
            self.assertEqual('ISCSI-target', target_name)
            return 'fake_map'

        mock_build_map.side_effect = build_map_func

        # Run the method
        self.vol_drv.connect_volume(self.slot_mgr)

        # As initialized above, remove_maps returns True to trigger update.
        self.assertEqual(1, mock_add_map.call_count)
        self.assertEqual(1, self.ft_fx.patchers['update'].mock.call_count)
        self.assertEqual(1, mock_build_map.call_count)
        mock_discover.assert_called_with(
            self.adpt, self.host_ip, self.user, self.password, self.iqn,
            self.feed[0].uuid, transport_type=self.trans_type)

    @mock.patch('pypowervm.tasks.hdisk.remove_iscsi', autospec=True)
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.hdisk_from_uuid',
                autospec=True)
    @mock.patch('pypowervm.tasks.scsi_mapper.remove_maps', autospec=True)
    @mock.patch('nova_powervm.virt.powervm.vm.get_vm_id')
    def test_disconnect_volume(self, mock_get_vm_id, mock_remove_maps,
                               mock_hdisk_from_uuid, mock_remove_iscsi):
        # The mock return values
        mock_hdisk_from_uuid.return_value = 'device_name'
        mock_get_vm_id.return_value = 'partition_id'
        self.vol_drv._set_devname('/dev/fake')

        def validate_remove_maps(vios_w, vm_uuid, match_func):
            self.assertIsInstance(vios_w, pvm_vios.VIOS)
            self.assertEqual('partition_id', vm_uuid)
            return 'removed'
        mock_remove_maps.side_effect = validate_remove_maps

        # Run the method
        self.vol_drv.disconnect_volume(self.slot_mgr)

        # As initialized above, remove_maps returns True to trigger update.
        self.assertEqual(1, mock_remove_maps.call_count)
        fake_iqn = self.vol_drv.connection_info['data']['target_iqn']
        mock_remove_iscsi.assert_called_once_with(
            self.adpt, fake_iqn, '3443DB77-AED1-47ED-9AA5-3DB9C6CF7089')
        self.assertEqual(1, self.ft_fx.patchers['update'].mock.call_count)

    def test_min_xags(self):
        xags = self.vol_drv.min_xags()
        self.assertEqual(1, len(xags))
        self.assertIn(pvm_const.XAG.VIO_SMAP, xags)

    def test_vol_type(self):
        self.assertEqual('iscsi', self.vol_drv.vol_type())

    def test_set_devname(self):

        # Mock connection info
        self.vol_drv.connection_info['data'][iscsi.DEVNAME_KEY] = None

        # Set the Device Name
        self.vol_drv._set_devname(self.devname)

        # Verify
        dev_name = self.vol_drv.connection_info['data'][iscsi.DEVNAME_KEY]
        self.assertEqual(self.devname, dev_name)

    def test_get_devname(self):

        # Set the value to retrieve
        self.vol_drv.connection_info['data'][iscsi.DEVNAME_KEY] = self.devname
        retrieved_devname = self.vol_drv._get_devname()
        # Check key found
        self.assertEqual(self.devname, retrieved_devname)

        # Check key not found
        self.vol_drv.connection_info['data'].pop(iscsi.DEVNAME_KEY)
        retrieved_devname = self.vol_drv._get_devname()
        # Check key not found
        self.assertIsNone(retrieved_devname)



# Copyright 2016, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
from nova import utils as n_utils
from oslo_concurrency import lockutils
from oslo_log import log as logging
from pypowervm import const as pvm_const
from pypowervm import exceptions as pvm_exc
import six
import time

from nova_powervm.virt.powervm.nvram import api
from nova_powervm.virt.powervm import vm

LOG = logging.getLogger(__name__)
LOCK_NVRAM_UPDT_LIST = 'nvram_update_list'
LOCK_NVRAM_STORE = 'nvram_update'


class NvramManager(object):
    """The manager of the NVRAM store and fetch process.

    This class uses two locks. One for controlling access to the list of
    instances to update the NVRAM for and another to control actually updating
    the NVRAM for the instance itself.

    An update to the instance store should always lock the update lock first
    and then get the list lock.  There should never be a case where the list
    lock is acquired before the update lock.  This can lead to deadlock cases.

    NVRAM events for an instance come in spurts primarily during power on and
    off, from what has been observed so far.  By using a dictionary and the
    instance.uuid as the key, rapid requests to store the NVRAM can be
    collapsed down into a single request (optimal).
    """

    def __init__(self, store_api, adapter, host_uuid):
        """Create the manager.

        :param store_api: the NvramStore api to use.
        :param adapter: pypowervm Adapter
        :param host_uuid:  powervm host uuid string
        """
        super(NvramManager, self).__init__()
        self._api = store_api
        self._adapter = adapter
        self._host_uuid = host_uuid

        self._update_list = {}
        self._queue = eventlet.queue.LightQueue()
        self._shutdown = False
        self._update_thread = n_utils.spawn(self._update_thread)
        LOG.debug('NVRAM store manager started.')

    def shutdown(self):
        """Shutdown the NVRAM Manager."""

        LOG.debug('NVRAM store manager shutting down.')
        self._shutdown = True
        # Remove all pending updates
        self._clear_list()
        # Signal the thread to stop
        self._queue.put(None)
        self._update_thread.wait()

    def store(self, instance, immediate=False):
        """Store the NVRAM for an instance.

        :param instance: The instance to store the NVRAM for.
        :param immediate: Force the update to take place immediately.
                          Otherwise, the request is queued for asynchronous
                          update.
        """
        if immediate:
            self._update_nvram(instance=instance)
        else:
            # Add it to the list to update
            self._add_to_list(instance)
            # Trigger the thread
            self._queue.put(instance.uuid, block=False)
            # Sleep so the thread gets a chance to run
            time.sleep(0)

    def fetch(self, instance):
        """Fetch the NVRAM for an instance.

        :param instance: The instance to fetch the NVRAM for.
        :returns: The NVRAM data for the instance.
        """
        try:
            return self._api.fetch(instance)
        except Exception as e:
            LOG.exception('Could not update NVRAM.', instance=instance)
            raise api.NVRAMDownloadException(instance=instance.name,
                                             reason=six.text_type(e))

    @lockutils.synchronized(LOCK_NVRAM_STORE)
    def remove(self, instance):
        """Remove the stored NVRAM for an instance.

        :param instance: The instance for which the NVRAM will be removed.
        """
        # Remove any pending updates
        self._pop_from_list(uuid=instance.uuid)
        # Remove it from the store
        try:
            self._api.delete(instance)
        except Exception:
            # Delete exceptions should not end the operation
            LOG.exception('Could not delete NVRAM.', instance=instance)

    @lockutils.synchronized(LOCK_NVRAM_UPDT_LIST)
    def _add_to_list(self, instance):
        """Add an instance to the list of instances to store the NVRAM."""
        self._update_list[instance.uuid] = instance

    @lockutils.synchronized(LOCK_NVRAM_UPDT_LIST)
    def _pop_from_list(self, uuid=None):
        """Pop an instance off the list of instance to update.

        :param uuid: The uuid of the instance to update or if not specified
                     pull the next instance off the list.
        returns: The uuid and instance.
        """
        try:
            if uuid is None:
                return self._update_list.popitem()
            else:
                return self._update_list.pop(uuid)
        except KeyError:
            return None, None

    @lockutils.synchronized(LOCK_NVRAM_UPDT_LIST)
    def _clear_list(self):
        """Clear the list of instance to store NVRAM for."""
        self._update_list.clear()

    @lockutils.synchronized(LOCK_NVRAM_STORE)
    def _update_nvram(self, instance=None):
        """Perform an update of NVRAM for instance.

        :param instance: The instance to update or if not specified pull the
                         next one off the list to update.
        """
        if instance is None:
            uuid, instance = self._pop_from_list()
            if uuid is None:
                return
        else:
            # Remove any pending updates
            self._pop_from_list(uuid=instance.uuid)

        try:
            LOG.debug('Updating NVRAM.', instance=instance)
            data = vm.get_instance_wrapper(
                self._adapter, instance, xag=[pvm_const.XAG.NVRAM]).nvram
            LOG.debug('NVRAM for instance: %s', data, instance=instance)
            if data is not None:
                self._api.store(instance, data)
        except pvm_exc.Error:
            # Update exceptions should not end the operation.
            LOG.exception('Could not update NVRAM.', instance=instance)

    def _update_thread(self):
        """The thread that is charged with updating the NVRAM store."""

        LOG.debug('NVRAM store manager update thread started.')
        # Loop until it's time to shut down
        while not self._shutdown:
            if self._queue.get(block=True) is None:
                LOG.debug('NVRAM store manager update thread is ending.')
                return

            self._update_nvram()
            time.sleep(0)



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from nova import test
from requests.exceptions import RequestException
from swiftclient import exceptions as swft_exc
from swiftclient import service as swft_srv

from nova_powervm.tests.virt import powervm
from nova_powervm.virt.powervm.nvram import api
from nova_powervm.virt.powervm.nvram import swift


class TestSwiftStore(test.TestCase):

    def setUp(self):
        super(TestSwiftStore, self).setUp()
        self.flags(swift_password='secret', swift_auth_url='url',
                   group='powervm')

        self.swift_store = swift.SwiftNvramStore()

    def test_run_operation(self):

        fake_result = [{'key1': 'value1'}, {'2key1', '2value1'}]
        fake_result2 = fake_result[0]

        def fake_generator(alist):
            for item in alist:
                yield item

        # Address the 'list' method that should be called.
        list_op = mock.Mock()
        self.swift_store.swift_service = mock.Mock(list=list_op)

        # Setup expected results
        list_op.return_value = fake_generator(fake_result)
        results = self.swift_store._run_operation('list', 1, x=2)

        list_op.assert_called_once_with(1, x=2)
        # Returns a copy of the results
        self.assertEqual(results, fake_result)
        self.assertNotEqual(id(results), id(fake_result))

        # Try a single result - Setup expected results
        list_op.reset_mock()
        list_op.return_value = fake_result2
        results = self.swift_store._run_operation('list', 3, x=4)

        list_op.assert_called_once_with(3, x=4)
        # Returns the actual result
        self.assertEqual(results, fake_result2)
        self.assertEqual(id(results), id(fake_result2))

        # Should raise any swift errors encountered
        list_op.side_effect = swft_srv.SwiftError('Error message.')
        self.assertRaises(swft_srv.SwiftError, self.swift_store._run_operation,
                          'list', 3, x=4)

    def _build_results(self, names):
        listing = [{'name': name} for name in names]
        return [{'success': True, 'listing': listing}]

    def test_get_name_from_listing(self):
        names = self.swift_store._get_name_from_listing(
            self._build_results(['snoopy']))
        self.assertEqual(['snoopy'], names)

    def test_get_container_names(self):
        with mock.patch.object(self.swift_store, '_run_operation') as mock_run:
            mock_run.return_value = self._build_results(['container'])
            names = self.swift_store._get_container_names()
            self.assertEqual(['container'], names)
            mock_run.assert_called_once_with('list',
                                             options={'long': True})

    @mock.patch('nova_powervm.virt.powervm.nvram.swift.SwiftNvramStore.'
                '_get_container_names')
    def test_get_object_names(self, mock_container_names):
        with mock.patch.object(self.swift_store, '_run_operation') as mock_run:
            mock_run.return_value = self._build_results(['obj', 'obj2'])

            # First run, no containers.
            mock_container_names.return_value = []
            names = self.swift_store._get_object_names('powervm_nvram')
            self.assertEqual([], names)
            self.assertEqual(1, mock_container_names.call_count)

            # Test without a prefix
            mock_container_names.return_value = ['powervm_nvram']
            names = self.swift_store._get_object_names('powervm_nvram')
            self.assertEqual(['obj', 'obj2'], names)
            mock_run.assert_called_once_with(
                'list', container='powervm_nvram',
                options={'long': True, 'prefix': None})
            self.assertEqual(mock_container_names.call_count, 2)

            # Test with a prefix
            names = self.swift_store._get_object_names('powervm_nvram',
                                                       prefix='obj')
            self.assertEqual(['obj', 'obj2'], names)
            mock_run.assert_called_with(
                'list', container='powervm_nvram',
                options={'long': True, 'prefix': 'obj'})

            # Second run should not increment the call count here
            self.assertEqual(mock_container_names.call_count, 2)

    @mock.patch('swiftclient.service.SwiftUploadObject')
    @mock.patch('nova_powervm.virt.powervm.nvram.swift.SwiftNvramStore.'
                '_exists')
    def test_underscore_store(self, mock_exists, mock_swiftuploadobj):
        mock_exists.return_value = True
        with mock.patch.object(self.swift_store, '_run_operation') as mock_run:
            mock_run.return_value = self._build_results(['obj'])
            self.swift_store._store(powervm.TEST_INST1.uuid,
                                    powervm.TEST_INST1.name, 'data')
            mock_run.assert_called_once_with('upload', 'powervm_nvram',
                                             mock.ANY, options=None)

            # Test unsuccessful upload
            mock_result = [{'success': False,
                            'error': RequestException('Error Message.')}]
            mock_run.return_value = mock_result
            self.assertRaises(api.NVRAMUploadException,
                              self.swift_store._store, powervm.TEST_INST1.uuid,
                              powervm.TEST_INST1.name, 'data')

            # Test retry upload
            mock_run.reset_mock()
            mock_swiftuploadobj.reset_mock()
            mock_res_obj = {'success': False,
                            'error': swft_exc.
                            ClientException('Error Message.'),
                            'object': '6ecb1386-53ab-43da-9e04-54e986ad4a9d'}
            mock_run.side_effect = [mock_res_obj,
                                    self._build_results(['obj'])]
            self.swift_store._store(powervm.TEST_INST1.uuid,
                                    powervm.TEST_INST1.name, 'data')
            mock_run.assert_called_with('upload', 'powervm_nvram',
                                        mock.ANY, options=None)
            self.assertEqual(mock_run.call_count, 2)
            self.assertEqual(mock_swiftuploadobj.call_count, 2)

    @mock.patch('swiftclient.service.SwiftUploadObject')
    @mock.patch('nova_powervm.virt.powervm.nvram.swift.SwiftNvramStore.'
                '_exists')
    def test_underscore_store_not_exists(self, mock_exists,
                                         mock_swiftuploadobj):
        mock_exists.return_value = False
        with mock.patch.object(self.swift_store, '_run_operation') as mock_run:
            mock_run.return_value = self._build_results(['obj'])
            self.swift_store._store(powervm.TEST_INST1.uuid,
                                    powervm.TEST_INST1.name, 'data')
            mock_run.assert_called_once_with(
                'upload', 'powervm_nvram', mock.ANY,
                options={'leave_segments': True})

            # Test retry upload
            mock_run.reset_mock()
            mock_swiftuploadobj.reset_mock()
            mock_res_obj = {'success': False,
                            'error': swft_exc.
                            ClientException('Error Message.'),
                            'object': '6ecb1386-53ab-43da-9e04-54e986ad4a9d'}
            mock_run.side_effect = [mock_res_obj,
                                    self._build_results(['obj'])]
            self.swift_store._store(powervm.TEST_INST1.uuid,
                                    powervm.TEST_INST1.name, 'data')
            mock_run.assert_called_with('upload', 'powervm_nvram', mock.ANY,
                                        options={'leave_segments': True})
            self.assertEqual(mock_run.call_count, 2)
            self.assertEqual(mock_swiftuploadobj.call_count, 2)

    @mock.patch('nova_powervm.virt.powervm.nvram.swift.SwiftNvramStore.'
                '_exists')
    def test_store(self, mock_exists):
        # Test forcing a update
        with mock.patch.object(self.swift_store, '_store') as mock_store:
            mock_exists.return_value = False
            self.swift_store.store(powervm.TEST_INST1, 'data', force=True)
            mock_store.assert_called_once_with(powervm.TEST_INST1.uuid,
                                               powervm.TEST_INST1.name,
                                               'data', exists=False)

        with mock.patch.object(
            self.swift_store, '_store') as mock_store, mock.patch.object(
                self.swift_store, '_run_operation') as mock_run:

            mock_exists.return_value = True
            data_md5_hash = '8d777f385d3dfec8815d20f7496026dc'
            results = self._build_results(['obj'])
            results[0]['headers'] = {'etag': data_md5_hash}
            mock_run.return_value = results
            self.swift_store.store(powervm.TEST_INST1, 'data', force=False)
            self.assertFalse(mock_store.called)
            mock_run.assert_called_once_with(
                'stat', options={'long': True},
                container='powervm_nvram', objects=[powervm.TEST_INST1.uuid])

    def test_store_slot_map(self):
        # Test forcing a update
        with mock.patch.object(self.swift_store, '_store') as mock_store:
            self.swift_store.store_slot_map("test_slot", 'data')
            mock_store.assert_called_once_with(
                'test_slot', 'test_slot', 'data')

    @mock.patch('os.remove')
    @mock.patch('tempfile.NamedTemporaryFile')
    @mock.patch('nova_powervm.virt.powervm.nvram.swift.SwiftNvramStore.'
                '_exists')
    def test_fetch(self, mock_exists, mock_tmpf, mock_rmv):
        mock_exists.return_value = True
        with mock.patch('nova_powervm.virt.powervm.nvram.swift.open',
                        mock.mock_open(read_data='data to read')
                        ) as m_open, mock.patch.object(
                self.swift_store, '_run_operation') as mock_run:
            mock_run.return_value = self._build_results(['obj'])
            mock_tmpf.return_value.__enter__.return_value.name = 'fname'

            data = self.swift_store.fetch(powervm.TEST_INST1)
            self.assertEqual('data to read', data)
            mock_rmv.assert_called_once_with(m_open.return_value.name)

            # Bad result from the download
            mock_run.return_value[0]['success'] = False
            self.assertRaises(api.NVRAMDownloadException,
                              self.swift_store.fetch, powervm.TEST_INST1)

    @mock.patch('os.remove')
    @mock.patch('tempfile.NamedTemporaryFile')
    @mock.patch('nova_powervm.virt.powervm.nvram.swift.SwiftNvramStore.'
                '_exists')
    def test_fetch_slot_map(self, mock_exists, mock_tmpf, mock_rmv):
        mock_exists.return_value = True
        with mock.patch('nova_powervm.virt.powervm.nvram.swift.open',
                        mock.mock_open(read_data='data to read')
                        ) as m_open, mock.patch.object(
                self.swift_store, '_run_operation') as mock_run:
            mock_run.return_value = self._build_results(['obj'])
            mock_tmpf.return_value.__enter__.return_value.name = 'fname'

            data = self.swift_store.fetch_slot_map("test_slot")
            self.assertEqual('data to read', data)
            mock_rmv.assert_called_once_with(m_open.return_value.name)

    @mock.patch('os.remove')
    @mock.patch('tempfile.NamedTemporaryFile')
    @mock.patch('nova_powervm.virt.powervm.nvram.swift.SwiftNvramStore.'
                '_exists')
    def test_fetch_slot_map_no_exist(self, mock_exists, mock_tmpf, mock_rmv):
        mock_exists.return_value = False
        data = self.swift_store.fetch_slot_map("test_slot")
        self.assertIsNone(data)

        # Make sure the remove (part of the finally block) is never called.
        # Should not get that far.
        self.assertFalse(mock_rmv.called)

    def test_delete(self):
        with mock.patch.object(self.swift_store, '_run_operation') as mock_run:
            mock_run.return_value = self._build_results(['obj'])
            self.swift_store.delete(powervm.TEST_INST1)
            mock_run.assert_called_once_with('delete',
                                             container='powervm_nvram',
                                             objects=[powervm.TEST_INST1.uuid])

            # Bad result from the operation
            mock_run.return_value[0]['success'] = False
            self.assertRaises(api.NVRAMDeleteException,
                              self.swift_store.delete, powervm.TEST_INST1)

    def test_delete_slot_map(self):
        with mock.patch.object(self.swift_store, '_run_operation') as mock_run:
            mock_run.return_value = self._build_results(['obj'])
            self.swift_store.delete_slot_map('test_slot')
            mock_run.assert_called_once_with(
                'delete', container='powervm_nvram', objects=['test_slot'])

            # Bad result from the operation
            mock_run.return_value[0]['success'] = False
            self.assertRaises(
                api.NVRAMDeleteException, self.swift_store.delete_slot_map,
                'test_slot')

    @mock.patch('nova_powervm.virt.powervm.nvram.swift.SwiftNvramStore.'
                '_get_object_names')
    def test_exists(self, mock_get_obj_names):
        # Test where there are elements in here
        mock_get_obj_names.return_value = ['obj', 'obj1', 'obj2']
        self.assertTrue(self.swift_store._exists('obj'))

        # Test where there are objects that start with the prefix, but aren't
        # actually there themselves
        mock_get_obj_names.return_value = ['obj1', 'obj2']
        self.assertFalse(self.swift_store._exists('obj'))

    def test_optional_options(self):
        """Test optional config values."""
        # Not in the sparse one from setUp()
        self.assertIsNone(self.swift_store.options['os_cacert'])
        self.assertIsNone(self.swift_store.options['os_endpoint_type'])
        # Create a new one with the optional values set
        self.flags(swift_cacert='/path/to/ca.pem', group='powervm')
        self.flags(swift_endpoint_type='internalURL', group='powervm')
        swift_store = swift.SwiftNvramStore()
        self.assertEqual('/path/to/ca.pem', swift_store.options['os_cacert'])
        self.assertEqual('internalURL',
                         swift_store.options['os_endpoint_type'])



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import mock

from nova import test
from nova_powervm.virt.powervm import exception as p_exc
from nova_powervm.virt.powervm import slot

from pypowervm import exceptions as pvm_exc


class TestNovaSlotManager(test.TestCase):

    def setUp(self):
        super(TestNovaSlotManager, self).setUp()
        self.store_api = mock.MagicMock()
        self.inst = mock.MagicMock(uuid='uuid1')

    def test_build_slot_mgr(self):
        # Test when NVRAM store exists
        # The Swift-backed implementation of PowerVM SlotMapStore is returned
        self.store_api.fetch_slot_map = mock.MagicMock(return_value=None)
        slot_mgr = slot.build_slot_mgr(self.inst, self.store_api, adapter=None,
                                       vol_drv_iter=None)
        self.assertIsInstance(slot_mgr, slot.SwiftSlotManager)
        self.assertFalse(slot_mgr.is_rebuild)

        # Test when no NVRAM store is set up
        # The no-op implementation of PowerVM SlotMapStore is returned
        self.assertIsInstance(
            slot.build_slot_mgr(self.inst, None, adapter=None,
                                vol_drv_iter=None),
            slot.NoopSlotManager)

        # Test that the rebuild flag is set when it is flagged as a rebuild
        slot_mgr = slot.build_slot_mgr(
            self.inst, self.store_api, adapter='adpt', vol_drv_iter='test')
        self.assertTrue(slot_mgr.is_rebuild)


class TestSwiftSlotManager(test.TestCase):

    def setUp(self):
        super(TestSwiftSlotManager, self).setUp()
        self.store_api = mock.MagicMock()
        self.store_api.fetch_slot_map = mock.MagicMock(return_value=None)
        self.inst = mock.MagicMock(uuid='a2e71b38-160f-4650-bbdc-2a10cd507e2b')
        self.slot_mgr = slot.SwiftSlotManager(self.store_api,
                                              instance=self.inst)

    def test_load(self):
        # load() should have been called internally by __init__
        self.store_api.fetch_slot_map.assert_called_with(
            self.inst.uuid + '_slot_map')

    def test_save(self):
        # Mock the call
        self.store_api.store_slot_map = mock.MagicMock()

        # Run save
        self.slot_mgr.save()

        # Not called because nothing changed
        self.store_api.store_slot_map.assert_not_called()

        # Change something
        mock_vfcmap = mock.Mock(server_adapter=mock.Mock(lpar_slot_num=123))
        self.slot_mgr.register_vfc_mapping(mock_vfcmap, 'fabric')

        # Run save
        self.slot_mgr.save()

        # Validate the call
        self.store_api.store_slot_map.assert_called_once_with(
            self.inst.uuid + '_slot_map', mock.ANY)

    def test_delete(self):
        # Mock the call
        self.store_api.delete_slot_map = mock.MagicMock()

        # Run delete
        self.slot_mgr.delete()

        # Validate the call
        self.store_api.delete_slot_map.assert_called_once_with(
            self.inst.uuid + '_slot_map')

    @mock.patch('pypowervm.tasks.slot_map.RebuildSlotMap')
    @mock.patch('pypowervm.tasks.storage.ComprehensiveScrub')
    def test_init_recreate_map(self, mock_ftsk, mock_rebuild_slot):
        vios1, vios2 = mock.Mock(uuid='uuid1'), mock.Mock(uuid='uuid2')
        mock_ftsk.return_value.feed = [vios1, vios2]
        self.slot_mgr.init_recreate_map(mock.Mock(), self._vol_drv_iter())
        self.assertEqual(1, mock_ftsk.call_count)
        mock_rebuild_slot.assert_called_once_with(
            self.slot_mgr, mock.ANY, {'udid': ['uuid2']}, ['a', 'b'])

    @mock.patch('pypowervm.tasks.slot_map.RebuildSlotMap')
    @mock.patch('pypowervm.tasks.storage.ComprehensiveScrub')
    def test_init_recreate_map_fails(self, mock_ftsk, mock_rebuild_slot):
        vios1, vios2 = mock.Mock(uuid='uuid1'), mock.Mock(uuid='uuid2')
        mock_ftsk.return_value.feed = [vios1, vios2]
        mock_rebuild_slot.side_effect = (
            pvm_exc.InvalidHostForRebuildNotEnoughVIOS(udid='udid56'))
        self.assertRaises(
            p_exc.InvalidRebuild, self.slot_mgr.init_recreate_map, mock.Mock(),
            self._vol_drv_iter())

    @mock.patch('pypowervm.tasks.slot_map.RebuildSlotMap')
    @mock.patch('pypowervm.tasks.storage.ComprehensiveScrub')
    def test_init_recreate_map_fileio(self, mock_ftsk, mock_rebuild_slot):
        vios1, vios2 = mock.Mock(uuid='uuid1'), mock.Mock(uuid='uuid2')
        mock_ftsk.return_value.feed = [vios1, vios2]
        expected_vio_wrap = [vios1, vios2]
        self.slot_mgr.init_recreate_map(mock.Mock(), self._vol_drv_iter_2())
        self.assertEqual(1, mock_ftsk.call_count)
        mock_rebuild_slot.assert_called_once_with(
            self.slot_mgr, expected_vio_wrap,
            {'udidvscsi': ['uuid1'], 'udid': ['uuid1']}, [])

    def _vol_drv_iter_2(self):
        mock_fileio = mock.Mock()
        mock_fileio.vol_type.return_value = 'fileio'
        mock_fileio.is_volume_on_vios.side_effect = ((True, 'udid'),
                                                     (False, None))
        mock_scsi = mock.Mock()
        mock_scsi.vol_type.return_value = 'vscsi'
        mock_scsi.is_volume_on_vios.side_effect = ((True, 'udidvscsi'),
                                                   (False, None))

        vol_drv = [mock_fileio, mock_scsi]
        for type in vol_drv:
            yield mock.Mock(), type

    def _vol_drv_iter(self):
        mock_scsi = mock.Mock()
        mock_scsi.vol_type.return_value = 'vscsi'
        mock_scsi.is_volume_on_vios.side_effect = ((False, None),
                                                   (True, 'udid'))

        mock_npiv1 = mock.Mock()
        mock_npiv1.vol_type.return_value = 'npiv'
        mock_npiv1._fabric_names.return_value = ['a', 'b']

        mock_npiv2 = mock.Mock()
        mock_npiv2.vol_type.return_value = 'npiv'
        mock_npiv2._fabric_names.return_value = ['a', 'b', 'c']

        vol_drv = [mock_scsi, mock_npiv1, mock_npiv2]
        for type in vol_drv:
            yield mock.Mock(), type



# Copyright 2014, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.compute import power_state
from nova.compute import task_states
from nova.compute import vm_states
from nova.objects import flavor
from nova.objects import image_meta
from nova.objects import instance
import os
import sys


TEST_FLAVOR = flavor.Flavor(memory_mb=2048,
                            swap=0,
                            vcpu_weight=None,
                            root_gb=10,
                            id=2,
                            name=u'm1.small',
                            ephemeral_gb=0,
                            rxtx_factor=1.0,
                            flavorid=u'1',
                            vcpus=1)

TEST_INSTANCE = {
    'id': 1,
    'uuid': '49629a5c-f4c4-4721-9511-9725786ff2e5',
    'display_name': 'Fake Instance',
    'root_gb': 10,
    'ephemeral_gb': 0,
    'instance_type_id': '5',
    'system_metadata': {'image_os_distro': 'rhel'},
    'host': 'host1',
    'flavor': TEST_FLAVOR,
    'task_state': None,
    'vm_state': vm_states.ACTIVE,
    'power_state': power_state.SHUTDOWN,
}

TEST_INST_SPAWNING = dict(TEST_INSTANCE, task_state=task_states.SPAWNING,
                          uuid='b3c04455-a435-499d-ac81-371d2a2d334f')

TEST_INST1 = instance.Instance(**TEST_INSTANCE)
TEST_INST2 = instance.Instance(**TEST_INST_SPAWNING)

TEST_MIGRATION = {
    'id': 1,
    'source_compute': 'host1',
    'dest_compute': 'host2',
    'migration_type': 'resize',
    'old_instance_type_id': 1,
    'new_instance_type_id': 2,
}
TEST_MIGRATION_SAME_HOST = dict(TEST_MIGRATION, dest_compute='host1')

IMAGE1 = {
    'id': '3e865d14-8c1e-4615-b73f-f78eaecabfbd',
    'name': 'image1',
    'size': 300,
    'container_format': 'bare',
    'disk_format': 'raw',
    'checksum': 'b518a8ba2b152b5607aceb5703fac072',
}
TEST_IMAGE1 = image_meta.ImageMeta.from_dict(IMAGE1)
EMPTY_IMAGE = image_meta.ImageMeta.from_dict({})

# NOTE(mikal): All of this is because if dnspython is present in your
# environment then eventlet monkeypatches socket.getaddrinfo() with an
# implementation which doesn't work for IPv6. What we're checking here is
# that the magic environment variable was set when the import happened.
if ('eventlet' in sys.modules and
        os.environ.get('EVENTLET_NO_GREENDNS', '').lower() != 'yes'):
    raise ImportError('eventlet imported before nova/cmd/__init__ '
                      '(env var set to %s)'
                      % os.environ.get('EVENTLET_NO_GREENDNS'))

os.environ['EVENTLET_NO_GREENDNS'] = 'yes'

import eventlet

eventlet.monkey_patch(os=False)



# Copyright 2015, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log as logging
from taskflow import task

from nova_powervm import conf as cfg
from nova_powervm.virt.powervm import exception as p_exc
from nova_powervm.virt.powervm.i18n import _
from nova_powervm.virt.powervm import vm

from pypowervm import const as pvm_const
from pypowervm.tasks import client_storage as pvm_c_stor
from pypowervm.tasks import hdisk
from pypowervm.tasks import scsi_mapper as tsk_map
from pypowervm.utils import transaction as tx
from pypowervm.wrappers import storage as pvm_stor
from pypowervm.wrappers import virtual_io_server as pvm_vios


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
UDID_KEY = 'target_UDID'
DEVNAME_KEY = 'target_devname'


class VscsiVolumeAdapter(object):
    """VscsiVolumeAdapter that connects a Cinder volume to a VM.

    This volume adapter is a generic adapter for volume types that use PowerVM
    vSCSI to host the volume to the VM.
    """

    def _connect_volume(self, slot_mgr):
        """Connects the volume.

        :param connect_volume_to_vio: Function to connect a volume to the vio.
                                      :param vios_w: Vios wrapper.
                                      :return: True if mapping was created.
        :param slot_mgr: A NovaSlotManager.  Used to delete the client slots
                         used when a volume is detached from the VM
        """

        # Its about to get weird.  The transaction manager has a list of
        # VIOSes.  We could use those, but they only have SCSI mappings (by
        # design).  They do not have storage (super expensive).
        #
        # We need the storage xag when we are determining which mappings to
        # add to the system.  But we don't want to tie it to the stg_ftsk.  If
        # we do, every retry, every etag gather, etc... takes MUCH longer.
        #
        # So we get the VIOSes with the storage xag here, separately, to save
        # the stg_ftsk from potentially having to run it multiple times.
        connect_ftsk = tx.FeedTask(
            'connect_volume_to_vio', pvm_vios.VIOS.getter(
                self.adapter, xag=[pvm_const.XAG.VIO_STOR,
                                   pvm_const.XAG.VIO_SMAP]))

        # Find valid hdisks and map to VM.
        connect_ftsk.add_functor_subtask(
            self._connect_volume_to_vio, slot_mgr, provides='vio_modified',
            flag_update=False)

        ret = connect_ftsk.execute()

        # Check the number of VIOSes
        vioses_modified = 0
        for result in ret['wrapper_task_rets'].values():
            if result['vio_modified']:
                vioses_modified += 1

        partition_id = vm.get_vm_id(self.adapter, self.vm_uuid)

        # Update the slot information
        def set_slot_info():
            vios_wraps = self.stg_ftsk.feed
            for vios_w in vios_wraps:
                scsi_map = pvm_c_stor.udid_to_scsi_mapping(
                    vios_w, self._get_udid(), partition_id)
                if not scsi_map:
                    continue
                slot_mgr.register_vscsi_mapping(scsi_map)

        self._validate_vios_on_connection(vioses_modified)
        self.stg_ftsk.add_post_execute(task.FunctorTask(
            set_slot_info, name='hdisk_slot_%s' % self._get_udid()))

    def _validate_vios_on_connection(self, num_vioses_found):
        """Validates that the correct number of VIOSes were discovered.

        Certain environments may have redundancy requirements.  For PowerVM
        this is achieved by having multiple Virtual I/O Servers.  This method
        will check to ensure that the operator's requirements for redundancy
        have been met.  If not, a specific error message will be raised.

        :param num_vioses_found: The number of VIOSes the hdisk was found on.
        """
        # Is valid as long as the vios count exceeds the conf value.
        if num_vioses_found >= CONF.powervm.vscsi_vios_connections_required:
            return

        # Should have a custom message based on zero or 'some but not enough'
        # I/O Servers.
        if num_vioses_found == 0:
            msg = (_('Failed to discover valid hdisk on any Virtual I/O '
                     'Server for volume %(volume_id)s.') %
                   {'volume_id': self.volume_id})
        else:
            msg = (_('Failed to discover the hdisk on the required number of '
                     'Virtual I/O Servers.  Volume %(volume_id)s required '
                     '%(vios_req)d Virtual I/O Servers, but the disk was only '
                     'found on %(vios_act)d Virtual I/O Servers.') %
                   {'volume_id': self.volume_id, 'vios_act': num_vioses_found,
                    'vios_req': CONF.powervm.vscsi_vios_connections_required})
        ex_args = {'volume_id': self.volume_id, 'reason': msg,
                   'instance_name': self.instance.name}
        raise p_exc.VolumeAttachFailed(**ex_args)

    def _add_append_mapping(self, vios_uuid, device_name, lpar_slot_num=None,
                            lua=None, target_name=None, udid=None):
        """Update the stg_ftsk to append the mapping to the VIOS.

        :param vios_uuid: The UUID of the vios for the pypowervm adapter.
        :param device_name: The The hdisk device name.
        :param lpar_slot_num: (Optional, Default:None) If specified, the client
                              lpar slot number to use on the mapping.  If left
                              as None, it will use the next available slot
                              number.
        :param lua: (Optional.  Default: None) Logical Unit Address to set on
                    the TargetDevice.  If None, the LUA will be assigned by the
                    server.  Should be specified for all of the VSCSIMappings
                    for a particular bus, or none of them.
        """
        def add_func(vios_w):
            LOG.info("Adding vSCSI mapping to Physical Volume %(dev)s",
                     {'dev': device_name}, instance=self.instance)
            pv = pvm_stor.PV.bld(self.adapter, device_name, udid)
            v_map = tsk_map.build_vscsi_mapping(
                self.host_uuid, vios_w, self.vm_uuid, pv,
                lpar_slot_num=lpar_slot_num, lua=lua, target_name=target_name)
            return tsk_map.add_map(vios_w, v_map)
        self.stg_ftsk.wrapper_tasks[vios_uuid].add_functor_subtask(add_func)

    def _get_udid(self):
        """This method will return the hdisk udid stored in connection_info.

        :return: The target_udid associated with the hdisk
        """
        try:
            return self.connection_info['data'][UDID_KEY]
        except (KeyError, ValueError):
            # It's common to lose our specific data in the BDM.  The connection
            # information can be 'refreshed' by operations like LPM and resize
            LOG.info('Failed to retrieve device_id key from BDM for volume id '
                     '%s', self.volume_id, instance=self.instance)
            return None

    def _set_udid(self, udid):
        """This method will set the hdisk udid in the connection_info.

        :param udid: The hdisk target_udid to be stored in system_metadata
        """
        self.connection_info['data'][UDID_KEY] = udid

    def _get_devname(self):
        """This method will return the hdisk devname stored in connection_info.

        :return: The target_devname associated with the hdisk
        """
        try:
            return self.connection_info['data'][DEVNAME_KEY]
        except (KeyError, ValueError):
            # It's common to lose our specific data in the BDM.  The connection
            # information can be 'refreshed' by operations like LPM and resize
            LOG.info('Failed to retrieve device_id key from BDM for volume id '
                     '%s', self.volume_id, instance=self.instance)
            return None

    def _set_devname(self, devname):
        """This method will set the hdisk devname in the connection_info.

        :param devname: The hdisk target_devname to be stored in
                        system_metadata
        """
        self.connection_info['data'][DEVNAME_KEY] = devname

    def _add_remove_mapping(self, vm_uuid, vios_uuid, device_name, slot_mgr):
        """Adds a transaction to remove the storage mapping.

        :param vm_uuid: The UUID of the VM instance
        :param vios_uuid: The UUID of the vios for the pypowervm adapter.
        :param device_name: The The hdisk device name.
        :param slot_mgr: A NovaSlotManager.  Used to delete the client slots
                         used when a volume is detached from the VM.
        """
        def rm_func(vios_w):
            LOG.info("Removing vSCSI mapping from physical volume %(dev)s.",
                     {'dev': device_name}, instance=self.instance)
            removed_maps = tsk_map.remove_maps(
                vios_w, vm_uuid,
                tsk_map.gen_match_func(pvm_stor.PV, names=[device_name]))
            for rm_map in removed_maps:
                slot_mgr.drop_vscsi_mapping(rm_map)
            return removed_maps
        self.stg_ftsk.wrapper_tasks[vios_uuid].add_functor_subtask(rm_func)

    def _add_remove_hdisk(self, vio_wrap, device_name,
                          stg_ftsk=None):
        """Adds a post-mapping task to remove the hdisk from the VIOS.

        This removal is only done after the mapping updates have completed.
        This method is also used during migration to remove hdisks that remain
        on the source host after the VM is migrated to the destination.

        :param vio_wrap: The Virtual I/O Server wrapper to remove the disk
                         from.
        :param device_name: The hdisk name to remove.
        :param stg_ftsk: The feed task to add to. If None, then self.stg_ftsk
        """
        def rm_hdisk():
            LOG.info("Removing hdisk %(hdisk)s from Virtual I/O Server "
                     "%(vios)s", {'hdisk': device_name, 'vios': vio_wrap.name},
                     instance=self.instance)
            try:
                # Attempt to remove the hDisk
                hdisk.remove_hdisk(self.adapter, CONF.host, device_name,
                                   vio_wrap.uuid)
            except Exception:
                # If there is a failure, log it, but don't stop the process
                LOG.exception("There was an error removing the hdisk "
                              "%(disk)s from Virtual I/O Server %(vios)s.",
                              {'disk': device_name, 'vios': vio_wrap.name},
                              instance=self.instance)

        # Check if there are not multiple mapping for the device
        if not self._check_host_mappings(vio_wrap, device_name):
            name = 'rm_hdisk_%s_%s' % (vio_wrap.name, device_name)
            stg_ftsk = stg_ftsk or self.stg_ftsk
            stg_ftsk.add_post_execute(task.FunctorTask(rm_hdisk, name=name))
        else:
            LOG.info("hdisk %(disk)s is not removed from Virtual I/O Server "
                     "%(vios)s because it has existing storage mappings",
                     {'disk': device_name, 'vios': vio_wrap.name},
                     instance=self.instance)

    def _check_host_mappings(self, vios_wrap, device_name):
        """Checks if the given hdisk has multiple mappings

        :param vio_wrap: The Virtual I/O Server wrapper to remove the disk
                         from.
        :param device_name: The hdisk name to remove.

        :return: True is there are multiple instances using the given hdisk
        """
        vios_scsi_mappings = next(v.scsi_mappings for v in self.stg_ftsk.feed
                                  if v.uuid == vios_wrap.uuid)
        mappings = tsk_map.find_maps(
            vios_scsi_mappings, None,
            tsk_map.gen_match_func(pvm_stor.PV, names=[device_name]))

        LOG.info("%(num)d storage mappings found for %(dev)s on VIOS %(vios)s",
                 {'num': len(mappings), 'dev': device_name,
                  'vios': vios_wrap.name}, instance=self.instance)
        # the mapping is still present as the task feed removes
        # the mapping later
        return len(mappings) > 1

    def _cleanup_volume(self, udid=None, devname=None):
        """Cleanup the hdisk associated with this udid."""

        if not udid and not devname:
            LOG.warning('Could not remove hdisk for volume %s', self.volume_id,
                        instance=self.instance)
            return

        LOG.info('Removing hdisk for udid: %s', udid, instance=self.instance)

        def find_hdisk_to_remove(vios_w):
            if devname is None:
                device_name = vios_w.hdisk_from_uuid(udid)
            else:
                device_name = devname
            if device_name is None:
                return
            LOG.info('Adding deferred task to remove %(hdisk)s from VIOS '
                     '%(vios)s.', {'hdisk': device_name, 'vios': vios_w.name},
                     instance=self.instance)
            self._add_remove_hdisk(vios_w, device_name,
                                   stg_ftsk=rmv_hdisk_ftsk)

        # Create a feed task to get the vios, find the hdisk and remove it.
        rmv_hdisk_ftsk = tx.FeedTask(
            'find_hdisk_to_remove', pvm_vios.VIOS.getter(
                self.adapter, xag=[pvm_const.XAG.VIO_STOR]))
        # Find vios hdisks for this udid to remove.
        rmv_hdisk_ftsk.add_functor_subtask(
            find_hdisk_to_remove, flag_update=False)
        rmv_hdisk_ftsk.execute()



# Copyright 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import fixtures
import mock

from nova import test
from pypowervm.wrappers import storage as pvm_stor
from pypowervm.wrappers import virtual_io_server as pvm_vios

from nova_powervm.virt.powervm.disk import imagecache as imagecache


class TestImageCache(test.TestCase):
    """Unit Tests for the LocalDisk storage driver."""

    def setUp(self):
        super(TestImageCache, self).setUp()
        self.mock_vg = mock.MagicMock(virtual_disks=[])
        # Initialize the ImageManager
        self.adpt = mock.MagicMock()
        self.vg_uuid = 'vg_uuid'
        self.vios_uuid = 'vios_uuid'
        self.img_cache = imagecache.ImageManager(self.vios_uuid, self.vg_uuid,
                                                 self.adpt)

        # Setup virtual_disks to be used later
        self.inst1 = pvm_stor.VDisk.bld(None, 'b_inst1', 10)
        self.inst2 = pvm_stor.VDisk.bld(None, 'b_inst2', 10)
        self.image = pvm_stor.VDisk.bld(None, 'i_bf8446e4_4f52', 10)

    def test_get_base(self):
        self.mock_vg_get = self.useFixture(fixtures.MockPatch(
            'pypowervm.wrappers.storage.VG.get')).mock
        self.mock_vg_get.return_value = self.mock_vg
        vg_wrap = self.img_cache._get_base()
        self.assertEqual(vg_wrap, self.mock_vg)
        self.mock_vg_get.assert_called_once_with(
            self.adpt, uuid=self.vg_uuid,
            parent_type=pvm_vios.VIOS.schema_type, parent_uuid=self.vios_uuid)

    def test_scan_base_image(self):
        # No cached images
        self.mock_vg.virtual_disks = [self.inst1, self.inst2]
        base_images = self.img_cache._scan_base_image(self.mock_vg)
        self.assertEqual([], base_images)
        # One 'cached' image
        self.mock_vg.virtual_disks.append(self.image)
        base_images = self.img_cache._scan_base_image(self.mock_vg)
        self.assertEqual([self.image], base_images)

    @mock.patch('pypowervm.tasks.storage.rm_vg_storage')
    @mock.patch('nova.virt.imagecache.ImageCacheManager.'
                '_list_running_instances')
    @mock.patch('nova_powervm.virt.powervm.disk.imagecache.ImageManager.'
                '_scan_base_image')
    def test_age_and_verify(self, mock_scan, mock_list, mock_rm):
        mock_context = mock.MagicMock()
        all_inst = mock.MagicMock()
        mock_scan.return_value = [self.image]
        # Two instances backed by image 'bf8446e4_4f52'
        # Mock dict returned from _list_running_instances
        used_images = {'': [self.inst1, self.inst2],
                       'bf8446e4_4f52': [self.inst1, self.inst2]}
        mock_list.return_value = {'used_images': used_images}

        self.mock_vg.virtual_disks = [self.inst1, self.inst2, self.image]
        self.img_cache._age_and_verify_cached_images(mock_context, all_inst,
                                                     self.mock_vg)
        mock_rm.assert_not_called()
        mock_scan.assert_called_once_with(self.mock_vg)
        mock_rm.reset_mock()

        # No instances
        mock_list.return_value = {'used_images': {}}
        self.img_cache._age_and_verify_cached_images(mock_context, all_inst,
                                                     self.mock_vg)
        mock_rm.assert_called_once_with(self.mock_vg, vdisks=[self.image])

    @mock.patch('nova_powervm.virt.powervm.disk.imagecache.ImageManager.'
                '_get_base')
    @mock.patch('nova_powervm.virt.powervm.disk.imagecache.ImageManager.'
                '_age_and_verify_cached_images')
    def test_update(self, mock_age, mock_base):
        mock_base.return_value = self.mock_vg
        mock_context = mock.MagicMock()
        mock_all_inst = mock.MagicMock()
        self.img_cache.update(mock_context, mock_all_inst)
        mock_base.assert_called_once_with()
        mock_age.assert_called_once_with(mock_context, mock_all_inst,
                                         self.mock_vg)



# Copyright 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from nova_powervm.tests.virt.powervm.volume import test_driver as test_vol
from nova_powervm.virt.powervm.volume import gpfs as v_drv


class TestGPFSVolumeAdapter(test_vol.TestVolumeAdapter):
    """Tests the GPFSVolumeAdapter.  NovaLink is a I/O host."""

    def setUp(self):
        super(TestGPFSVolumeAdapter, self).setUp()

        # Needed for the volume adapter
        self.adpt = mock.Mock()
        mock_inst = mock.MagicMock(uuid='2BC123')

        # Connection Info
        mock_conn_info = {'data': {'device_path': '/gpfs/path'}}

        self.vol_drv = v_drv.GPFSVolumeAdapter(
            self.adpt, 'host_uuid', mock_inst, mock_conn_info)

    def test_get_path(self):
        self.assertEqual('/gpfs/path', self.vol_drv._get_path())



# Copyright 2015, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import absolute_import

import fixtures
import mock

from nova import exception
from nova import objects
from nova.objects import migrate_data as mig_obj
from nova import test
from nova.tests.unit import fake_network

from nova_powervm.tests.virt import powervm
from nova_powervm.tests.virt.powervm import fixtures as fx
from nova_powervm.virt.powervm import live_migration as lpm


class TestLPM(test.TestCase):
    def setUp(self):
        super(TestLPM, self).setUp()

        self.flags(disk_driver='localdisk', group='powervm')
        self.drv_fix = self.useFixture(fx.PowerVMComputeDriver())
        self.drv = self.drv_fix.drv
        self.apt = self.drv.adapter

        self.inst = objects.Instance(**powervm.TEST_INSTANCE)

        self.network_infos = fake_network.fake_get_instance_nw_info(self, 1)
        self.inst.info_cache = objects.InstanceInfoCache(
            network_info=self.network_infos)

        self.mig_data = mig_obj.PowerVMLiveMigrateData()
        self.mig_data.host_mig_data = {}
        self.mig_data.dest_ip = '1'
        self.mig_data.dest_user_id = 'neo'
        self.mig_data.dest_sys_name = 'a'
        self.mig_data.public_key = 'PublicKey'
        self.mig_data.dest_proc_compat = 'a,b,c'
        self.mig_data.vol_data = {}
        self.mig_data.vea_vlan_mappings = {}

        self.lpmsrc = lpm.LiveMigrationSrc(self.drv, self.inst, self.mig_data)
        self.lpmdst = lpm.LiveMigrationDest(self.drv, self.inst)

        self.add_key = self.useFixture(fixtures.MockPatch(
            'pypowervm.tasks.management_console.add_authorized_key')).mock
        self.get_key = self.useFixture(fixtures.MockPatch(
            'pypowervm.tasks.management_console.get_public_key')).mock
        self.get_key.return_value = 'PublicKey'

        # Short path to the host's migration_data
        self.host_mig_data = self.drv.host_wrapper.migration_data

    @mock.patch('pypowervm.tasks.storage.ScrubOrphanStorageForLpar')
    @mock.patch('nova_powervm.virt.powervm.media.ConfigDrivePowerVM')
    @mock.patch('nova_powervm.virt.powervm.vm.get_instance_wrapper')
    @mock.patch('pypowervm.tasks.vterm.close_vterm')
    def test_lpm_source(self, mock_vterm_close, mock_get_wrap,
                        mock_cd, mock_scrub):
        self.host_mig_data['active_migrations_supported'] = 4
        self.host_mig_data['active_migrations_in_progress'] = 2

        with mock.patch.object(
            self.lpmsrc, '_check_migration_ready', return_value=None):

            # Test the bad path first, then patch in values to make succeed
            mock_wrap = mock.Mock(id=123)
            mock_get_wrap.return_value = mock_wrap

            self.assertRaises(exception.MigrationPreCheckError,
                              self.lpmsrc.check_source, 'context',
                              'block_device_info', [])

            # Patch the proc compat fields, to get further
            pm = mock.PropertyMock(return_value='b')
            type(mock_wrap).proc_compat_mode = pm

            self.assertRaises(exception.MigrationPreCheckError,
                              self.lpmsrc.check_source, 'context',
                              'block_device_info', [])

            pm = mock.PropertyMock(return_value='Not_Migrating')
            type(mock_wrap).migration_state = pm

            # Get a volume driver.
            mock_vol_drv = mock.MagicMock()

            # Finally, good path.
            self.lpmsrc.check_source('context', 'block_device_info',
                                     [mock_vol_drv])
            # Ensure we built a scrubber.
            mock_scrub.assert_called_with(mock.ANY, 123)
            # Ensure we added the subtasks to remove the vopts.
            mock_cd.return_value.dlt_vopt.assert_called_once_with(
                mock.ANY, stg_ftsk=mock_scrub.return_value,
                remove_mappings=False)
            # And ensure the scrubber was executed
            mock_scrub.return_value.execute.assert_called_once_with()
            mock_vol_drv.pre_live_migration_on_source.assert_called_once_with(
                {})

            # Ensure migration counts are validated
            self.host_mig_data['active_migrations_in_progress'] = 4
            self.assertRaises(exception.MigrationPreCheckError,
                              self.lpmsrc.check_source, 'context',
                              'block_device_info', [])

            # Ensure the vterm was closed
            mock_vterm_close.assert_called_once_with(
                self.apt, mock_wrap.uuid)

    def test_lpm_dest(self):
        src_compute_info = {'stats': {'memory_region_size': 1}}
        dst_compute_info = {'stats': {'memory_region_size': 1}}

        self.host_mig_data['active_migrations_supported'] = 4
        self.host_mig_data['active_migrations_in_progress'] = 2
        with mock.patch.object(self.drv.host_wrapper, 'refresh') as mock_rfh:

            self.lpmdst.check_destination(
                'context', src_compute_info, dst_compute_info)
            mock_rfh.assert_called_once_with()

            # Ensure migration counts are validated
            self.host_mig_data['active_migrations_in_progress'] = 4
            self.assertRaises(exception.MigrationPreCheckError,
                              self.lpmdst.check_destination, 'context',
                              src_compute_info, dst_compute_info)
            # Repair the stat
            self.host_mig_data['active_migrations_in_progress'] = 2

            # Ensure diff memory sizes raises an exception
            dst_compute_info['stats']['memory_region_size'] = 2
            self.assertRaises(exception.MigrationPreCheckError,
                              self.lpmdst.check_destination, 'context',
                              src_compute_info, dst_compute_info)

    @mock.patch('pypowervm.tasks.storage.ComprehensiveScrub')
    @mock.patch('nova_powervm.virt.powervm.vif.'
                'pre_live_migrate_at_destination')
    def test_pre_live_mig(self, mock_vif_pre, mock_scrub):
        vol_drv = mock.MagicMock()
        network_infos = [{'type': 'pvm_sea'}]

        def update_vea_mapping(adapter, host_uuid, instance, network_info,
                               vea_vlan_mappings):
            # Make sure what comes in is None, but that we change it.
            self.assertEqual(vea_vlan_mappings, {})
            vea_vlan_mappings['test'] = 'resp'

        mock_vif_pre.side_effect = update_vea_mapping

        resp = self.lpmdst.pre_live_migration(
            'context', 'block_device_info', network_infos, 'disk_info',
            self.mig_data, [vol_drv])

        # Make sure the pre_live_migrate_at_destination was invoked for the vif
        mock_vif_pre.assert_called_once_with(
            self.drv.adapter, self.drv.host_uuid, self.inst, network_infos[0],
            mock.ANY)
        self.assertEqual({'test': 'resp'}, self.mig_data.vea_vlan_mappings)

        # Make sure we get something back, and that the volume driver was
        # invoked.
        self.assertIsNotNone(resp)
        vol_drv.pre_live_migration_on_destination.assert_called_once_with(
            self.mig_data.vol_data)
        self.assertEqual(1, mock_scrub.call_count)
        self.add_key.assert_called_once_with(self.apt, 'PublicKey')

        vol_drv.reset_mock()
        raising_vol_drv = mock.Mock()
        raising_vol_drv.pre_live_migration_on_destination.side_effect = (
            Exception('foo'))
        self.assertRaises(
            exception.MigrationPreCheckError, self.lpmdst.pre_live_migration,
            'context', 'block_device_info', network_infos, 'disk_info',
            self.mig_data, [vol_drv, raising_vol_drv])
        vol_drv.pre_live_migration_on_destination.assert_called_once_with({})
        (raising_vol_drv.pre_live_migration_on_destination.
            assert_called_once_with({}))

    def test_src_cleanup(self):
        vol_drv = mock.Mock()
        self.lpmdst.cleanup_volume(vol_drv)
        # Ensure the volume driver is not called
        self.assertEqual(0, vol_drv.cleanup_volume_at_destination.call_count)

    def test_src_cleanup_valid(self):
        vol_drv = mock.Mock()
        self.lpmdst.pre_live_vol_data = {'vscsi-vol-id': 'fake_udid'}
        self.lpmdst.cleanup_volume(vol_drv)
        # Ensure the volume driver was called to clean up the volume.
        vol_drv.cleanup_volume_at_destination.assert_called_once()

    @mock.patch('pypowervm.tasks.migration.migrate_lpar')
    @mock.patch('nova_powervm.virt.powervm.live_migration.LiveMigrationSrc.'
                '_convert_nl_io_mappings')
    @mock.patch('nova_powervm.virt.powervm.vif.pre_live_migrate_at_source')
    def test_live_migration(self, mock_vif_pre_lpm, mock_convert_mappings,
                            mock_migr):
        mock_trunk = mock.MagicMock()
        mock_vif_pre_lpm.return_value = [mock_trunk]
        mock_convert_mappings.return_value = ['AABBCCDDEEFF/5']

        self.lpmsrc.lpar_w = mock.Mock()
        self.lpmsrc.live_migration('context', self.mig_data)
        mock_migr.assert_called_once_with(
            self.lpmsrc.lpar_w, 'a', sdn_override=True, tgt_mgmt_svr='1',
            tgt_mgmt_usr='neo', validate_only=False,
            virtual_fc_mappings=None, virtual_scsi_mappings=None,
            vlan_check_override=True, vlan_mappings=['AABBCCDDEEFF/5'])

        # Network assertions
        mock_vif_pre_lpm.assert_called_once_with(
            self.drv.adapter, self.drv.host_uuid, self.inst, mock.ANY)
        mock_trunk.delete.assert_called_once()

        # Test that we raise errors received during migration
        mock_migr.side_effect = ValueError()
        self.assertRaises(ValueError, self.lpmsrc.live_migration, 'context',
                          self.mig_data)
        mock_migr.assert_called_with(
            self.lpmsrc.lpar_w, 'a', sdn_override=True, tgt_mgmt_svr='1',
            tgt_mgmt_usr='neo', validate_only=False,
            virtual_fc_mappings=None, virtual_scsi_mappings=None,
            vlan_mappings=['AABBCCDDEEFF/5'], vlan_check_override=True)

    def test_convert_nl_io_mappings(self):
        # Test simple None case
        self.assertIsNone(self.lpmsrc._convert_nl_io_mappings(None))

        # Do some mappings
        test_mappings = {'aa:bb:cc:dd:ee:ff': 5, 'aa:bb:cc:dd:ee:ee': 126}
        expected = ['AABBCCDDEEFF/5', 'AABBCCDDEEEE/126']
        self.assertEqual(
            set(expected),
            set(self.lpmsrc._convert_nl_io_mappings(test_mappings)))

    @mock.patch('pypowervm.tasks.migration.migrate_recover')
    def test_rollback(self, mock_migr):
        self.lpmsrc.lpar_w = mock.Mock()

        # Test no need to rollback
        self.lpmsrc.lpar_w.migration_state = 'Not_Migrating'
        self.lpmsrc.rollback_live_migration('context')
        self.assertTrue(self.lpmsrc.lpar_w.refresh.called)
        self.assertFalse(mock_migr.called)

        # Test calling the rollback
        self.lpmsrc.lpar_w.reset_mock()
        self.lpmsrc.lpar_w.migration_state = 'Pretend its Migrating'
        self.lpmsrc.rollback_live_migration('context')
        self.assertTrue(self.lpmsrc.lpar_w.refresh.called)
        mock_migr.assert_called_once_with(self.lpmsrc.lpar_w, force=True)

        # Test exception from rollback
        mock_migr.reset_mock()
        self.lpmsrc.lpar_w.reset_mock()
        mock_migr.side_effect = ValueError()
        self.lpmsrc.rollback_live_migration('context')
        self.assertTrue(self.lpmsrc.lpar_w.refresh.called)
        mock_migr.assert_called_once_with(self.lpmsrc.lpar_w, force=True)

    def test_check_migration_ready(self):
        lpar_w, host_w = mock.Mock(), mock.Mock()
        lpar_w.can_lpm.return_value = (True, None)
        self.lpmsrc._check_migration_ready(lpar_w, host_w)
        lpar_w.can_lpm.assert_called_once_with(host_w, migr_data={})

        lpar_w.can_lpm.return_value = (False, 'This is the reason message.')
        self.assertRaises(exception.MigrationPreCheckError,
                          self.lpmsrc._check_migration_ready, lpar_w, host_w)

    @mock.patch('pypowervm.tasks.migration.migrate_abort')
    def test_migration_abort(self, mock_mig_abort):
        self.lpmsrc.lpar_w = mock.Mock()
        self.lpmsrc.migration_abort()
        mock_mig_abort.called_once_with(self.lpmsrc.lpar_w)

    @mock.patch('pypowervm.tasks.migration.migrate_recover')
    def test_migration_recover(self, mock_mig_recover):
        self.lpmsrc.lpar_w = mock.Mock()
        self.lpmsrc.migration_recover()
        mock_mig_recover.called_once_with(self.lpmsrc.lpar_w, force=True)

    @mock.patch('nova_powervm.virt.powervm.vif.post_live_migrate_at_source')
    def test_post_live_migration_at_source(self, mock_vif_post_lpm_at_source):
        network_infos = [{'devname': 'tap-dev1', 'address': 'mac-addr1',
                          'network': {'bridge': 'br-int'}, 'id': 'vif_id_1'},
                         {'devname': 'tap-dev2', 'address': 'mac-addr2',
                          'network': {'bridge': 'br-int'}, 'id': 'vif_id_2'}]
        self.lpmsrc.post_live_migration_at_source(network_infos)
        # Assertions
        for network_info in network_infos:
            mock_vif_post_lpm_at_source.assert_any_call(mock.ANY, mock.ANY,
                                                        mock.ANY, network_info)



# Copyright 2015, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from nova.api.metadata import base as instance_metadata
from nova.network import model as network_model
from nova.virt import configdrive
import os
import retrying
from taskflow import task

from oslo_log import log as logging
from oslo_utils import excutils

from pypowervm import const as pvm_const
from pypowervm.tasks import scsi_mapper as tsk_map
from pypowervm.tasks import storage as tsk_stg
from pypowervm.tasks import vopt as tsk_vopt
from pypowervm import util as pvm_util
from pypowervm.utils import transaction as pvm_tx
from pypowervm.wrappers import storage as pvm_stg
from pypowervm.wrappers import virtual_io_server as pvm_vios

from nova_powervm import conf as cfg
from nova_powervm.virt.powervm import vm


LOG = logging.getLogger(__name__)

CONF = cfg.CONF

_LLA_SUBNET = "fe80::/64"

CFG_DRV_PREFIX = "cfg_"
CFG_DRV_SUFFIX = ".iso"


class ConfigDrivePowerVM(object):

    def __init__(self, adapter, host_uuid):
        """Creates the config drive manager for PowerVM.

        :param adapter: The pypowervm adapter to communicate with the system.
        :param host_uuid: The UUID of the host system.
        """
        self.adapter = adapter
        self.host_uuid = host_uuid

        # Validate that the virtual optical exists
        self.vios_uuid, self.vg_uuid = tsk_vopt.validate_vopt_repo_exists(
            self.adapter,
            vopt_media_volume_group=CONF.powervm.vopt_media_volume_group,
            vopt_media_rep_size=CONF.powervm.vopt_media_rep_size)

    @staticmethod
    def _sanitize_network_info(network_info):
        """Will sanitize the network info for the config drive.

        Newer versions of cloud-init look at the vif type information in
        the network info and utilize it to determine what to do.  There are
        a limited number of vif types, and it seems to be built on the idea
        that the neutron vif type is the cloud init vif type (which is not
        quite right).

        This sanitizes the network info that gets passed into the config
        drive to work properly with cloud-inits.
        """
        network_info = copy.deepcopy(network_info)
        for vif in network_info:
            if vif.get('type') is not 'ovs':
                vif['type'] = 'vif'
        return network_info

    def _create_cfg_dr_iso(self, instance, injected_files, network_info,
                           admin_pass=None):
        """Creates an ISO file that contains the injected files.

        Used for config drive.

        :param instance: The VM instance from OpenStack.
        :param injected_files: A list of file paths that will be injected into
                               the ISO.
        :param network_info: The network_info from the nova spawn method.
        :param admin_pass: Optional password to inject for the VM.
        :return iso_path: The path to the ISO
        :return file_name: The file name for the ISO
        """
        LOG.info("Creating config drive.", instance=instance)
        extra_md = {}
        if admin_pass is not None:
            extra_md['admin_pass'] = admin_pass

        # Sanitize the vifs for the network config
        network_info = self._sanitize_network_info(network_info)

        inst_md = instance_metadata.InstanceMetadata(instance,
                                                     content=injected_files,
                                                     extra_md=extra_md,
                                                     network_info=network_info)

        # Make sure the path exists.
        im_path = CONF.powervm.image_meta_local_path
        if not os.path.exists(im_path):
            os.mkdir(im_path)

        file_name = pvm_util.sanitize_file_name_for_api(
            instance.name, prefix=CFG_DRV_PREFIX, suffix=CFG_DRV_SUFFIX,
            max_len=pvm_const.MaxLen.VOPT_NAME)
        iso_path = os.path.join(im_path, file_name)
        with configdrive.ConfigDriveBuilder(instance_md=inst_md) as cdb:
            LOG.info("Config drive ISO building to path %(iso_path)s.",
                     {'iso_path': iso_path}, instance=instance)
            # In case, if there's an OSError related failure while
            # creating config drive, retry make drive operation.

            def _retry_on_oserror(exc):
                return isinstance(exc, OSError)

            @retrying.retry(retry_on_exception=_retry_on_oserror,
                            stop_max_attempt_number=2)
            def _make_cfg_drive(iso_path):
                cdb.make_drive(iso_path)
            try:
                _make_cfg_drive(iso_path)
                return iso_path, file_name
            except OSError:
                with excutils.save_and_reraise_exception(logger=LOG):
                    # If we get here, that means there's an exception during
                    # second attempt, log the same and fail the deploy
                    # operation.
                    LOG.exception("Config drive ISO could not be built.",
                                  instance=instance)

    def create_cfg_drv_vopt(self, instance, injected_files, network_info,
                            lpar_uuid, admin_pass=None, mgmt_cna=None,
                            stg_ftsk=None):
        """Creates the config drive virtual optical and attach to VM.

        :param instance: The VM instance from OpenStack.
        :param injected_files: A list of file paths that will be injected into
                               the ISO.
        :param network_info: The network_info from the nova spawn method.
        :param lpar_uuid: The UUID of the client LPAR
        :param admin_pass: (Optional) password to inject for the VM.
        :param mgmt_cna: (Optional) The management (RMC) CNA wrapper.
        :param stg_ftsk: (Optional) If provided, the tasks to create and attach
                         the Media to the VM will be deferred on to the
                         FeedTask passed in.  The execute can be done all in
                         one method (batched together).  If None (the default),
                         the media will be created and attached immediately.
        """
        # If there is a management client network adapter, then we should
        # convert that to a VIF and add it to the network info
        if mgmt_cna is not None and CONF.powervm.use_rmc_ipv6_scheme:
            network_info = copy.deepcopy(network_info)
            network_info.append(self._mgmt_cna_to_vif(mgmt_cna))

        iso_path, file_name = self._create_cfg_dr_iso(instance, injected_files,
                                                      network_info, admin_pass)

        # Upload the media
        file_size = os.path.getsize(iso_path)
        vopt, f_uuid = self._upload_vopt(iso_path, file_name, file_size)

        # Delete the media
        os.remove(iso_path)

        # Run the attach of the virtual optical
        self._attach_vopt(instance, lpar_uuid, vopt, stg_ftsk)

    def _attach_vopt(self, instance, lpar_uuid, vopt, stg_ftsk=None):
        """Will attach the vopt to the VIOS.

        If the stg_ftsk is provided, adds the mapping to the stg_ftsk, but
        won't attach until the stg_ftsk is independently executed.

        :param instance: The VM instance from OpenStack.
        :param lpar_uuid: The UUID of the client LPAR
        :param vopt: The virtual optical device to add.
        :param stg_ftsk: (Optional) If provided, the tasks to create the
                         storage mappings to connect the Media to the VM will
                         be deferred on to the FeedTask passed in.  The execute
                         can be done all in one method (batched together).  If
                         None (the default), the media will be attached
                         immediately.
        """
        # If no transaction manager, build locally so that we can run
        # immediately
        if stg_ftsk is None:
            wtsk = pvm_tx.WrapperTask('media_attach', pvm_vios.VIOS.getter(
                self.adapter, entry_uuid=self.vios_uuid,
                xag=[pvm_const.XAG.VIO_SMAP]))
        else:
            wtsk = stg_ftsk.wrapper_tasks[self.vios_uuid]

        # Define the function to build and add the mapping
        def add_func(vios_w):
            LOG.info("Adding config drive mapping to Virtual I/O Server "
                     "%(vios)s", {'vios': vios_w.name}, instance=instance)
            mapping = tsk_map.build_vscsi_mapping(self.host_uuid, vios_w,
                                                  lpar_uuid, vopt)
            return tsk_map.add_map(vios_w, mapping)

        wtsk.add_functor_subtask(add_func)

        # If built locally, then execute
        if stg_ftsk is None:
            wtsk.execute()

    def _mgmt_cna_to_vif(self, cna):
        """Converts the mgmt CNA to VIF format for network injection."""
        # See IEFT RFC 4291 appendix A for information on this algorithm
        mac = vm.norm_mac(cna.mac)
        ipv6_link_local = self._mac_to_link_local(mac)

        subnet = network_model.Subnet(
            version=6, cidr=_LLA_SUBNET,
            ips=[network_model.FixedIP(address=ipv6_link_local)])
        network = network_model.Network(id='mgmt', subnets=[subnet],
                                        injected='yes')
        return network_model.VIF(id='mgmt_vif', address=mac,
                                 network=network)

    @staticmethod
    def _mac_to_link_local(mac):
        # Convert the address to IPv6.  The first step is to separate out the
        # mac address
        splits = mac.split(':')

        # Insert into the middle the key ff:fe
        splits.insert(3, 'ff')
        splits.insert(4, 'fe')

        # Do the bit flip on the first octet.
        splits[0] = "%.2x" % (int(splits[0], 16) ^ 0b00000010)

        # Convert to the IPv6 link local format.  The prefix is fe80::.  Join
        # the hexes together at every other digit.
        ll = ['fe80:']
        ll.extend([splits[x] + splits[x + 1]
                   for x in range(0, len(splits), 2)])
        return ':'.join(ll)

    def _upload_vopt(self, iso_path, file_name, file_size):
        with open(iso_path, 'rb') as d_stream:
            return tsk_stg.upload_vopt(self.adapter, self.vios_uuid, d_stream,
                                       file_name, file_size)

    def dlt_vopt(self, lpar_uuid, stg_ftsk=None, remove_mappings=True):
        """Deletes the virtual optical and scsi mappings for a VM.

        :param lpar_uuid: The pypowervm UUID of the LPAR to remove.
        :param stg_ftsk: (Optional) A FeedTask. If provided, the actions to
                         modify the storage will be added as batched functions
                         onto the FeedTask.  If not provided (the default) the
                         operation to delete the vOpt will execute immediately.
        :param remove_mappings: (Optional, Default: True) If set to true, will
                                remove the SCSI mappings as part of the
                                operation.  If false, will leave the mapping
                                but detach the storage from it.  If the VM is
                                running, it may be necessary to do the latter
                                as some operating systems will not allow the
                                removal.
        """
        # If no transaction manager, build locally so that we can run
        # immediately
        if stg_ftsk is None:
            built_stg_ftsk = True
            vio_w = pvm_vios.VIOS.get(self.adapter, root_id=self.vios_uuid,
                                      xag=[pvm_const.XAG.VIO_SMAP])
            stg_ftsk = pvm_tx.FeedTask('media_detach', [vio_w])
        else:
            built_stg_ftsk = False

        # Run the remove maps method.
        self.add_dlt_vopt_tasks(lpar_uuid, stg_ftsk,
                                remove_mappings=remove_mappings)

        # If built locally, then execute
        if built_stg_ftsk:
            stg_ftsk.execute()

    def add_dlt_vopt_tasks(self, lpar_uuid, stg_ftsk, remove_mappings=True):
        """Deletes the virtual optical and (optionally) scsi mappings for a VM.

        :param lpar_uuid: The pypowervm UUID of the LPAR whose vopt is to be
                          removed.
        :param stg_ftsk: A FeedTask handling storage I/O.  The task to remove
                         the mappings and media from the VM will be deferred on
                         to the FeedTask passed in. The execute can be done all
                         in one method (batched together).  No updates are
                         actually made here; they are simply added to the
                         FeedTask.
        :param remove_mappings: (Optional, Default: True) If set to true, will
                                remove the SCSI mappings as part of the
                                operation.  If false, will leave the mapping
                                but detach the storage from it.  If the VM is
                                running, it may be necessary to do the latter
                                as some operating systems will not allow the
                                removal.
        """
        # The function to find the VOpt
        match_func = tsk_map.gen_match_func(pvm_stg.VOptMedia)

        def rm_vopt_mapping(vios_w):
            return tsk_map.remove_maps(vios_w, lpar_uuid,
                                       match_func=match_func)

        def detach_vopt_from_map(vios_w):
            return tsk_map.detach_storage(vios_w, lpar_uuid,
                                          match_func=match_func)

        # Add a function to remove the map or detach the vopt
        stg_ftsk.wrapper_tasks[self.vios_uuid].add_functor_subtask(
            rm_vopt_mapping if remove_mappings else detach_vopt_from_map)

        # Find the vOpt device (before the remove is done) so that it can be
        # removed.
        partition_id = vm.get_vm_id(self.adapter, lpar_uuid)
        media_mappings = tsk_map.find_maps(
            stg_ftsk.get_wrapper(self.vios_uuid).scsi_mappings,
            client_lpar_id=partition_id, match_func=match_func)
        media_elems = [x.backing_storage for x in media_mappings]

        def rm_vopt():
            LOG.info("Removing virtual optical for VM with UUID %s.",
                     lpar_uuid)
            vg_wrap = pvm_stg.VG.get(self.adapter, uuid=self.vg_uuid,
                                     parent_type=pvm_vios.VIOS,
                                     parent_uuid=self.vios_uuid)
            tsk_stg.rm_vg_storage(vg_wrap, vopts=media_elems)

        # Don't add this task if there is no media to delete (eg. config drive)
        if media_elems:
            stg_ftsk.add_post_execute(task.FunctorTask(rm_vopt))
